#pragma once

#include <stdbool.h>
#include <math.h>
#include <stdio.h>

bool is_close(float a, float b)
{
  const float rtol = 1e-8f;
  const float atol = 1e-5f;
  return fabsf(a - b) <= (atol + rtol * fabsf(b));
}

#define RESET   "\033[0m"
#define BLACK   "\033[30m"      /* Black */
#define RED     "\033[31m"      /* Red */
#define GREEN   "\033[32m"      /* Green */
#define YELLOW  "\033[33m"      /* Yellow */
#define BLUE    "\033[34m"      /* Blue */
#define MAGENTA "\033[35m"      /* Magenta */
#define CYAN    "\033[36m"      /* Cyan */
#define WHITE   "\033[37m"      /* White */
#define BOLDBLACK   "\033[1m\033[30m"      /* Bold Black */
#define BOLDRED     "\033[1m\033[31m"      /* Bold Red */
#define BOLDGREEN   "\033[1m\033[32m"      /* Bold Green */
#define BOLDYELLOW  "\033[1m\033[33m"      /* Bold Yellow */
#define BOLDBLUE    "\033[1m\033[34m"      /* Bold Blue */
#define BOLDMAGENTA "\033[1m\033[35m"      /* Bold Magenta */
#define BOLDCYAN    "\033[1m\033[36m"      /* Bold Cyan */
#define BOLDWHITE   "\033[1m\033[37m"      /* Bold White */

#define PASS() do {\
    fprintf(stderr, BOLDGREEN "%s passed\n" RESET, __func__);\
  } while(0)

#define EXPECT(cond) do {\
    if(!(cond))\
    {\
      fprintf(stderr, BOLDRED "%s failed in line %d\n" RESET, __func__, __LINE__);\
    }\
  } while(0)

#define EXPECT_CLOSE(a, b) do {\
  if(!is_close((a), (b)))\
  {\
    fprintf(stderr, BOLDRED "%s failed in line %d: expected a = %g, got b = %g\n" RESET, __func__, __LINE__, a, b);\
    return;\
  }\
  } while(0)

#define EXPECT_ALL_CLOSE(a, b, n) do {\
    for(int i = 0; i < (n); i++) {\
      if(!is_close(a[i], b[i])) {\
        fprintf(stderr, BOLDRED "%s failed in line %d: expected a[%d] = %g, got b[%d] = %g\n" RESET, __func__, __LINE__, i, a[i], i, b[i]);\
        return;\
      }\
    }\
  } while(0)


