#pragma once

// expectation maximization with:
// - modified pca-part initialization https://ece.northeastern.edu/fac-ece/jdy/papers/initialization-su-dy-ida07.pdf
// - and mixture reduction https://ieeexplore.ieee.org/document/4383588

#include "config.h"

#include <stdint.h>
#include <stdlib.h>
#include "heap.h"

#include <stdint.h>

typedef struct guided_stats_t guided_stats_t;

typedef struct gmm_comp_accum_t
{
  double weight;
  double weight2; // sum of squared weights
  double mean[GUIDE_DIM];
  double cov[(GUIDE_DIM * (GUIDE_DIM+1)) / 2]; // upper triangular matrix
} gmm_comp_accum_t;

typedef struct gmm_fit_comp_t
{
  float weight;
  float mean[GUIDE_DIM];
  float E_T[GUIDE_DIM * GUIDE_DIM];
  float w_inv[GUIDE_DIM];
#if 0
  gmm_comp_accum_t prior;
#endif
} gmm_fit_comp_t;

typedef struct gmm_fit_t
{
  uint32_t n_components;
  uint32_t components_allocated;
  gmm_fit_comp_t *components;
  gmm_comp_accum_t *accums;
  void *scratch;
  heap_t *heap;
} gmm_fit_t;

typedef struct em_sample_t
{
  float weight;
  float x[GUIDE_DIM];
} em_sample_t;

void gmm_fit_init(gmm_fit_t *gmm);
void gmm_fit_cleanup(gmm_fit_t *gmm);
void gmm_fit_resize(gmm_fit_t *gmm, uint32_t n_components);

void gmm_fit_samples(gmm_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius, guided_stats_t *stats);
void gmm_fit_reduce(gmm_fit_t *gmm, float threshold, double total_weight);

//--- camera mixture -----------------------------------------------------------

typedef struct gmm3_comp_accum_t
{
  double weight;
  double weight2; // sum of squared weights
  double mean[3];
  double cov[(3 * (3+1)) / 2]; // upper triangular matrix
} gmm3_comp_accum_t;

typedef struct gmm3_fit_comp_t
{
  float weight;
  float mean[3];
  float E_T[3 * 3];
  float w_inv[3];
} gmm3_fit_comp_t;

typedef struct gmm3_fit_t
{
  uint32_t n_components;
  uint32_t components_allocated;
  gmm3_fit_comp_t *components;
  gmm3_comp_accum_t *accums;
  void *scratch;
  heap_t *heap;
} gmm3_fit_t;

void gmm3_fit_init(gmm3_fit_t *gmm);
void gmm3_fit_cleanup(gmm3_fit_t *gmm);
void gmm3_fit_resize(gmm3_fit_t *gmm, uint32_t n_components);

void gmm3_copy_from_full_gmm(gmm3_fit_t *gmm, const gmm_fit_t *full_gmm);

void gmm3_fit_samples(gmm3_fit_t *gmm, em_sample_t *samples, uint32_t n_samples, float prior_radius);
void gmm3_fit_reduce(gmm3_fit_t *gmm, float threshold, double total_weight);

uint32_t em_sample_filter_aabb(em_sample_t *samples, uint32_t n_samples, const float *cell_aabb);
