#pragma once

#include "pathspace.h"

#include <stdbool.h>
#include <stdint.h>

typedef struct guided_t guided_t;
typedef struct pixel_est_t pixel_est_t;

typedef enum render_mode_t
{
  rm_init,
  rm_learn,
  rm_render,
} render_mode_t;

guided_t *guided_init();
void guided_cleanup(guided_t *c);

void guided_update(guided_t *c, pixel_est_t *px_est, int num_samples);

void guided_prepare_frame(guided_t *c);

int guided_extend(guided_t *c, path_t *path);

md_t guided_pdf_extend_dwp(guided_t *c, path_t *path, int v);

void guided_record_path(guided_t *c, path_t *path, double throughput, render_mode_t render_mode);

md_t guided_eval_pdf_pt(guided_t *c, path_t *path);
md_t guided_eval_pdf_lt(guided_t *c, path_t *path);

// stats
void guided_print_stats(guided_t *c, FILE *f, int json);
void guided_reset_stats(guided_t *c);

// cache serialization
void guided_dump_cache(const guided_t *c, const char *filename);
int guided_load_cache(guided_t *c, const char *filename);
void guided_dump_cache_debug(guided_t *c, const char *filename);

float guided_path_mis_uniform(guided_t *c, path_t *path, int v);

void guided_successful_path(guided_t *c, path_t *path);

void guided_path_print(guided_t *c, path_t *path);

void guided_new_path(guided_t *c);

void guided_invalidate_lobe_list(guided_t *g, int new_length, bool lt);

void guided_reuse_lobe_list(guided_t *g, int new_length, bool lt);
