#include "lobe_list.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

static void lobe_list_realloc(lobe_list_t *ll);

void lobe_list_init(lobe_list_t *ll, int n)
{
  ll->lobes = NULL;
  ll->cdf = NULL;
  ll->allocated = n;
  lobe_list_realloc(ll);
  lobe_list_clear(ll);
}

void lobe_list_cleanup(lobe_list_t *ll)
{
  free(ll->lobes);
  free(ll->cdf);
}

void lobe_list_extend(lobe_list_t *ll)
{
  ll->allocated *= 2;
  lobe_list_realloc(ll);
}

void lobe_list_clear(lobe_list_t *ll)
{
  ll->end = 0;
  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
    lobe_list_clear_entry(ll, i);
  ll->last_path_length = 0;
}

int lobe_list_get_length(const lobe_list_t *ll, int v)
{
  return ll->ranges[v].end - ll->ranges[v].begin;
}

static void lobe_list_realloc(lobe_list_t *ll)
{
  const size_t n = ll->allocated;
  ll->lobes = realloc(ll->lobes, n * sizeof(*ll->lobes));
  ll->cdf = realloc(ll->cdf, n * sizeof(*ll->cdf));
}

void lobe_list_clear_entry(lobe_list_t *ll, int v)
{
  ll->ranges[v] = (lobe_range_t) { .begin = 0, .end = -1 };
}

void lobe_list_update_end(lobe_list_t *ll)
{
  int end = 0;
  for(int i = 0; i < PATHSPACE_MAX_VERTS; i++)
  {
    if(ll->ranges[i].end > end) end = ll->ranges[i].end;
  }
  ll->end = end;
}

void lobe_list_shift_new_path(lobe_list_t *ll, int new_path_length)
{
  assert(new_path_length >= ll->last_path_length);
  int shift = new_path_length - ll->last_path_length;
  memmove(ll->ranges + shift, ll->ranges, sizeof(lobe_range_t) * ll->last_path_length);
}
