#include "lobe_type.h"

#include "pathspace.h"

#include <assert.h>

lobe_type_t lobe_type_from_path(const path_t *path, int v)
{
  int lt = path->v[0].mode & s_emit;
  lobe_type_t type = 0;

  if(v > 0)
  {
    // these are only set on the path once we evaluated the bsdf
    if(path->v[v].mode & s_reflect) type |= t_reflect;
    else if(path->v[v].mode & s_transmit) type |= t_transmit;

    if(path->v[v].mode & s_volume) type |= t_vol_v2;

#if VERTEX_PAIRS == 0
    // also set flags for v1 in case of triplets
    if(lt)
    {
      if(path->v[v-1].flags & s_environment) type |= t_envmap;
      else if(path->v[v-1].mode & s_volume) type |= t_vol_v3;
      else type |= t_surf_v3;
    }
    else
    {
      if(path->v[v-1].mode & s_sensor) type |= t_camera;
      else if(path->v[v-1].mode & s_volume) type |= t_vol_v1;
      else type |= t_surf_v1;
    }
#endif
  }
  else assert(!lt);

  if(lt)
  {
    if(path->v[v+1].mode & s_volume) type |= t_vol_v1;
    else if(path->v[v+1].mode & s_sensor) type |= t_camera;
    else type |= t_surf_v1;
  }
  else
  {
    if(path->v[v+1].mode & s_volume) type |= t_vol_v3;
    else if(path->v[v+1].flags & s_environment) type |= t_envmap;
    else type |= t_surf_v3;
  }

  return type;
}

uint32_t lobe_type_get_cull_mask(lobe_type_t type)
{ // this really only applies to reflect/transmit so far
  if(type & t_reflect) return s_transmit;
  else if(type & t_transmit) return s_reflect;
  return 0;
}

bool lobe_type_is_supported(lobe_type_t type, const path_t *path, int v)
{
  int lt = path->v[0].mode & s_emit;
  if(v == 0)
  {
    assert(!lt);
    if(path->v[1].flags & s_environment) return false;
    bool path_in_volume = (path->v[1].mode & s_volume) != 0;
    bool lobe_in_volume = (type & t_vol_v3) != 0;
    return path_in_volume == lobe_in_volume;
  }
  else
  {
    bool supported = true;

#if VERTEX_PAIRS == 0
    vertex_scattermode_t prev_mode = path->v[v-1].mode;
    vertex_flags_t prev_flags = path->v[v-1].flags;
    lobe_type_t path_type;
    lobe_type_t mask;
    if(lt)
    {
      mask = t_surf_v3 | t_vol_v3 | t_envmap;
      if(prev_flags & s_environment) path_type = t_envmap;
      else if(prev_mode & s_volume) path_type = t_vol_v3;
      else path_type = t_surf_v3;
    }
    else
    {
      mask = t_camera | t_vol_v1 | t_surf_v1;
      if(prev_mode & s_sensor) path_type = t_camera;
      else if(prev_mode & s_volume) path_type = t_vol_v1;
      else path_type = t_surf_v1;
    }
    supported = (mask & type) == path_type;
#endif

    vertex_scattermode_t material_modes = path->v[v].material_modes;
    if(type & t_vol_v2)   supported &= (material_modes & s_volume)   != 0;
    if(type & t_reflect)  supported &= (material_modes & s_reflect)  != 0;
    if(type & t_transmit) supported &= (material_modes & s_transmit) != 0;
    return supported;
  }
}

void lobe_type_set_tentative(lobe_type_t type, path_t *path, int v)
{
  int lt = path->v[0].mode & s_emit;
  if(type & t_transmit) path->v[v-1].mode |= s_transmit;
  if(type & t_reflect) path->v[v-1].mode |= s_reflect;
  if(lt)
  {
    if(type & t_vol_v1) path->v[v].mode = s_volume;
  }
  else
  {
    if(type & t_vol_v3) path->v[v].mode = s_volume;
    if(type & t_envmap) path->v[v].flags = s_environment;
  }
}
