#pragma once

#include <stdbool.h>
#include <stdint.h>

typedef struct path_t path_t;

// v1, v2 and v3 are always in camera->light direction
typedef enum lobe_type_t
{ // only one bit of each group should be set
  // v1: camera, surface or volume
  t_camera   = 1 << 0,
  t_vol_v1   = 1 << 7,
  t_surf_v1  = 1 << 8,

  // v2: reflect, transmit or volume scatter
  t_reflect  = 1 << 1,
  t_transmit = 1 << 2,
  t_vol_v2   = 1 << 3,

  // v3: surface, volume or envmap
  t_surf_v3  = 1 << 4,
  t_vol_v3   = 1 << 5,
  t_envmap   = 1 << 6,
} lobe_type_t;

// maximum number of different lobe types
#define N_LOBE_TYPES (3*3*3)

// v is index of middle triplet vertex here
lobe_type_t lobe_type_from_path(const path_t *path, int v);

uint32_t lobe_type_get_cull_mask(lobe_type_t type);

// check if we can consider this lobe at the current vertex to extend this path
// v is index of middle triplet vertex (at which we want to extend)
bool lobe_type_is_supported(lobe_type_t type, const path_t *path, int v);

// sets env and vol flags for the vertex to be sampled (index v)
void lobe_type_set_tentative(lobe_type_t type, path_t *path, int v);
