#pragma once

#include "config.h"
#include "lobe_type.h"

#include <stdint.h>
#include <stdlib.h>
#include "heap.h"

// implementation of "Stream sampling for variance-optimal estimation of subset sums"
// https://arxiv.org/abs/0803.0473

typedef struct reservoir_sample_t
{
  float x[GUIDE_DIM];
  lobe_type_t type;
} reservoir_sample_t;

typedef struct reservoir_t
{
  heap_t *heap; // contains indices into beginning of samples array
  reservoir_sample_t *samples; // heap samples in front, equal-weight at back of array
  uint32_t samples_begin; // index where equal-weight samples begin in samples array
  uint32_t capacity;
  double equal_weight; // weight of each sample in the equal-weight region of samples
  uint32_t samples_seen;
} reservoir_t;

void reservoir_init(reservoir_t *reservoir, size_t capacity);

void reservoir_cleanup(reservoir_t *reservoir);

void reservoir_push(reservoir_t *reservoir, const reservoir_sample_t *sample, float weight, float xi[2]);

void reservoir_clear(reservoir_t *reservoir);

void reservoir_decay(reservoir_t *reservoir, float decay);

uint32_t reservoir_stored_samples(const reservoir_t *reservoir);
