#pragma once

#include <stdint.h>
#include <stdio.h>

typedef enum extend_error_t
{
  NO_ERROR,
  ERR_MAX_LENGTH,
  ERR_UNIFORM_EXTEND_FIRST,
  ERR_CAM_SAMPLE,
  ERR_UNIFORM_EXTEND,
  ERR_BSDF_SAMPLE,
  ERR_DIST_SAMPLE,
  ERR_FIRST_OUTSIDE_AABB,
  ERR_WRONG_MODE_V3,
  ERR_WRONG_FLAGS,
  ERR_BSDF_ZERO,
  ERR_ENV_ABSORB,
  ERR_WRONG_MODE_V2,
  ERR_PDF_ZERO_2D,
  ERR_PDF_ZERO_3D,
  ERR_NO_MATERIAL,

  NUM_EXTEND_ERRORS, // dummy for counting
}
extend_error_t;

typedef struct guided_stats_t guided_stats_t;

guided_stats_t *stats_init(void);
void stats_cleanup(guided_stats_t *stats);

void stats_collect(guided_stats_t *stats_);
void stats_reset(guided_stats_t *stats);

typedef struct guided_cache_stats_t
{
  int num_lobes_v1;
  int num_lobes_total;
}
guided_cache_stats_t;
void stats_print(guided_stats_t *stats, guided_cache_stats_t cache_stats, FILE *f);
void stats_print_json(guided_stats_t *stats, guided_cache_stats_t cache_stats, FILE *f);

void stats_record_lobes_found(guided_stats_t *stats, uint64_t n_lobes, int v);
void stats_record_lobes_filtered(guided_stats_t *stats, uint64_t n_lobes, int v);
void stats_record_lobes_used(guided_stats_t *stats, uint64_t n_lobes, int v);

void stats_record_err_extend(guided_stats_t *stats, extend_error_t err);
void stats_record_err_project(guided_stats_t *stats, int err);

void stats_record_successful_path(guided_stats_t *stats);
void stats_record_sampled_path(guided_stats_t *stats);

void stats_record_em(guided_stats_t *stats, double start, double after_init, double after_em, int iterations);
