#include "reservoir.c"
#include "unittest.h"

#include <stdlib.h>

static float randf(void)
{
  return (double) rand() / RAND_MAX;
}

static void test_init(void)
{
  reservoir_t reservoir;
  reservoir_init(&reservoir, 10);
  reservoir_cleanup(&reservoir);
  PASS();
}

static void test_push(void)
{
  reservoir_t reservoir;
  reservoir_init(&reservoir, 10);
  for(int i = 0; i < 100; i++)
  {
    reservoir_sample_t sample;
    float random[2] = { randf(), randf() };
    reservoir_push(&reservoir, &sample, 101. - i, random);
  }
  EXPECT(reservoir.heap->end + (reservoir.capacity + 1 - reservoir.samples_begin) == reservoir.capacity);
  reservoir_cleanup(&reservoir);
  PASS();
}

static void test_mean(float factor)
{
  double total = 0;
  double total_weight = 0;
  double total_reservoir = 0;
  double total_weight_reservoir = 0;
  uint32_t N = 1000;
  for(uint32_t i = 0; i < N; i++)
  {
    reservoir_t reservoir;
    reservoir_init(&reservoir, N * factor);
    for(uint32_t j = 0; j < N; j++)
    {
      float x = randf();
      float w = 2. + x + (randf() * 0.5f - 0.5f);
      total += x * w;
      total_weight += w;
      reservoir_sample_t s = { .x = { x } };
      float random[2] = { randf(), randf() };
      reservoir_push(&reservoir, &s, w, random);
    }
    heap_t *h = reservoir.heap;
    for(uint32_t j = 0; j < h->end; j++)
    {
      float w = -h->vals[j];
      total_reservoir += w * reservoir.samples[h->keys[j]].x[0];
      total_weight_reservoir += w;
    }
    for(uint32_t j = reservoir.samples_begin; j < reservoir.capacity + 1; j++)
    {
      float w = reservoir.equal_weight;
      total_reservoir += w * reservoir.samples[j].x[0];
      total_weight_reservoir += w;
    }
    reservoir_cleanup(&reservoir);
  }
  float mean = total / total_weight;
  float mean_reservoir = total_reservoir / total_weight_reservoir;
  printf("%g %g\n", mean, mean_reservoir);
}

static void test_mean_01(void)
{
  test_mean(0.1);
  PASS();
}

static void test_mean_05(void)
{
  test_mean(0.5);
  PASS();
}

static void test_mean_09(void)
{
  test_mean(0.9);
  PASS();
}

static void test_mean_10(void)
{
  test_mean(1.0);
  PASS();
}

int main(int argc, char *argv[])
{
  (void) argc, (void) argv;

  test_init();
  test_push();

  test_mean_01();
  test_mean_05();
  test_mean_09();
  test_mean_10();

  return 0;
}
