#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
namespace py = pybind11;

extern "C"
{
  #include "gmm.h"
  #include "guided.h"
}

void accum_set(gmm_lobe_accum_t *accum, py::array_t<double> array)
{
  py::buffer_info buffer = array.request();
  if (buffer.ndim != 1) throw std::runtime_error("Number of dimensions must be one");
  if (buffer.size != 55) throw std::runtime_error("Input size != 55");

  const double *ptr = (const double*) buffer.ptr;
  for(int i = 0; i < 55; i++)
  {
    accum->lobe[i] = ptr[i];
  }
}

py::array_t<double> accum_get(gmm_lobe_accum_t *accum)
{
  auto result = py::array_t<double>(55);
  py::buffer_info buffer = result.request();
  double *ptr = (double*) buffer.ptr;
  for(int i = 0; i < 55; i++)
  {
    ptr[i] = accum->lobe[i];
  }
  return result;
}

py::array_t<double> gmm_sample_pos3d_pybind(const gmm_lobe_t *lobe, py::array_t<double> random, py::array_t<double> xpp, py::array_t<double> xp)
{
  float random_arr[3];
  float xpp_arr[3];
  float xp_arr[3];
  float xn_arr[3];

  for(int k = 0; k < 3; k++)
  {
    random_arr[k] = random.at(k);
    xpp_arr[k] = xpp.at(k);
    xp_arr[k] = xp.at(k);
  }

  gmm_sample_pos3d(lobe, random_arr, xpp_arr, xp_arr, xn_arr);

  auto xn = py::array_t<double>(3);
  py::buffer_info buffer = xn.request();
  double *ptr = (double*) buffer.ptr;
  for(int k = 0; k < 3; k++)
  {
    ptr[k] = xn_arr[k];
  }
  return xn;
}

py::array_t<double> gmm_sample_pos2d_pybind(const gmm_lobe_t *lobe, py::array_t<double> random, py::array_t<double> xpp, py::array_t<double> xp)
{
  float random_arr[2];
  float xpp_arr[3];
  float xp_arr[3];
  float xn_arr[3];

  random_arr[0] = random.at(0);
  random_arr[1] = random.at(1);
  for(int k = 0; k < 3; k++)
  {
    xpp_arr[k] = xpp.at(k);
    xp_arr[k] = xp.at(k);
  }

  gmm_sample_pos2d(lobe, random_arr, xpp_arr, xp_arr, xn_arr);

  auto xn = py::array_t<double>(3);
  py::buffer_info buffer = xn.request();
  double *ptr = (double*) buffer.ptr;
  for(int k = 0; k < 3; k++)
  {
    ptr[k] = xn_arr[k];
  }
  return xn;
}

double gmm_sample_dist_pybind(const gmm_lobe_t *lobe, double random, py::array_t<double> xp, py::array_t<double> dir)
{
  float xp_arr[3];
  float dir_arr[3];

  for(int k = 0; k < 3; k++)
  {
    xp_arr[k] = xp.at(k);
    dir_arr[k] = dir.at(k);
  }

  float out_dist;
  gmm_sample_dist(lobe, random, xp_arr, dir_arr, &out_dist);
  return out_dist;
}

PYBIND11_MODULE(guided, m) {
  m.doc() = "";

  m.def("accum_set", &accum_set, "");
  m.def("accum_get", &accum_get, "");
  m.def("gmm_update_lobe", &gmm_update_lobe, "");
  m.def("gmm_sample_pos3d", &gmm_sample_pos3d_pybind, "");
  m.def("gmm_sample_pos2d", &gmm_sample_pos2d_pybind, "");
  m.def("gmm_sample_dist", &gmm_sample_dist_pybind, "");
  py::class_<gmm_lobe_accum_t>(m, "gmm_lobe_accum_t")
    .def(py::init());
  py::class_<gmm_lobe_t>(m, "gmm_lobe_t")
    .def(py::init());
}

// symbols need to be present but are not actually used
gmm_lobe_accum_t* guided_get_lobe_accum(guided_cache_t *c, uint64_t index)
{
  return NULL;
}

gmm_lobe_t* guided_get_lobe(guided_cache_t *c, uint64_t index)
{
  return NULL;
}
