# what to try:
# initialization:
# - random center points, assign to closest
# - existing lobes (only means or means + cov)
# - kmeans
# - kmeans++
# try with and without truncated gaussians
# export histogram of sample weights!

import bpy
from bpy import context as C
from bpy import data as D
from mathutils import *

import numpy as np
import struct
import copy
import heapq

import importlib
import lobe
importlib.reload(lobe)
import octree
importlib.reload(octree)
import lobevis
importlib.reload(lobevis)

from lobe import Lobe
from octree import Octree
from lobevis import LobeVis

scene_dir = '/home/vincent/scenes/guiding3v/'
scene = 'testpool'
postfix = 'render'
cache_iter = 64
guide_pairs = True
guide_dim = 9 if not guide_pairs else 6
#query_pos = np.array([0., 0., 0.])
query_pos = np.array([407., 138., -87.]) # end of pool floor
#query_pos = np.array([391., 384., -103.])
#query_pos = np.array([408., -36., -39.])
#query_pos = np.array([505.894, 337.771, -41.3406]) # left side of pool
em_iterations = 250

em_vis = None

def main(synthetic=False):
    if synthetic:
        grouped_samples = [synthetic_samples()]
        prior_radius = 0.1
    else:
        cell_aabb = load_cell_aabb(query_pos)
        side_length = np.amin(cell_aabb[3:] - cell_aabb[:3])
        prior_radius = 0.01 * side_length
        grouped_samples, configs = load_samples_filtered(cell_aabb)
    #reduction_steps = max(len(priors) for priors in grouped_priors)
    reduction_steps = 0 # TODO
    all_lobes = []
    for samples in grouped_samples:
        if len(samples.targets) == 0:
            continue
        #samples_subset = resample_varopt(samples, 0.5)
        mixture_all_lobes = em_fit(samples, prior_radius)
        for i, mixture in enumerate(mixture_all_lobes):
            new_lobes = [comp.to_lobe() for comp in mixture]
            if i < len(all_lobes):
                all_lobes[i].extend(new_lobes)
            else:
                all_lobes.append(new_lobes)
        """ TODO
        mixture = mixture_all_lobes[-1]
        acc_cost = 0
        for i in range(reduction_steps):
            mixture, cost = mixture_reduce_step(mixture)
            acc_cost += cost
            all_lobes[em_iterations + i].extend(mixture)
            print(acc_cost)
        """
    global em_vis
    em_vis = EMVis(all_lobes)
    bpy.app.timers.register(timer_callback)

def timer_callback():
    global em_vis
    time_to_next_call = 0.1
    if em_vis.update():
        return time_to_next_call
    else:
        return None
    return None

query_name = 'query'

class EMVis:
    def __init__(self, all_lobes):
        self.frame = None
        self.lobe_vis = LobeVis(all_lobes, show_all=True)
        self.make_query_object()

    def update(self):
        if query_name not in D.objects:
            print("stopping")
            return False
        query_obj = D.objects[query_name]
        pos_did_change = query_obj.location != self.query_pos
        new_frame = bpy.context.scene.frame_current
        frame_did_change = new_frame != self.frame
        self.frame = new_frame

        self.query_pos = Vector(query_obj.location)

        if frame_did_change or pos_did_change:
            self.lobe_vis.update(self.frame, np.array(self.query_pos))
        return True

    def make_query_object(self):
        if query_name in D.objects:
            D.objects.remove(D.objects[query_name])
        bpy.ops.object.empty_add(type='PLAIN_AXES')
        query_obj = C.active_object
        query_obj.name = query_name
        query_obj.empty_display_size = 10
        self.query_pos = None

def load_cell_aabb(query_pos):
    cache_path = (scene_dir
            + scene
            +'/corona/'
            + scene
            + postfix
            + '_cache{:05d}_vis'.format(cache_iter))
    with open(cache_path, 'rb') as f:
        Lobe.read_from_file(f) # could also just skip ahead in file, we only want the octree
        octree = Octree.read_from_file(f)
    cell_aabb = octree.lookup_aabb(query_pos)
    return cell_aabb

def inside_aabb(aabb, point):
    return np.all(point >= aabb[:3]) and np.all(point <= aabb[3:])

def inside_aabb_vec(aabb, point):
    return np.logical_and(np.all(point >= aabb[:3], axis=1), np.all(point <= aabb[3:], axis=1))

def load_samples_filtered(cell_aabb):
    samples_path = (scene_dir
            + scene
            +'/corona/'
            + scene
            + postfix
            + '_cache{:05d}_samples'.format(cache_iter))
    samples = []
    sample_format = struct.Struct(
        '='
        + 'f' # target
        + 'I' # mode v2, v3 and flags
        + '9f' # vertex triplet
    ) if not guide_pairs else struct.Struct(
        '='
        + 'f' # target
        + 'I' # mode v2, v3 and flags
        + '6f' # vertex pair
    )
    grouped_samples = []
    v2off = 0 if guide_pairs else 3
    with open(samples_path, 'rb') as f:
        guide_dim_file, = struct.unpack('i', f.read(4))
        assert(guide_dim_file == guide_dim)
        n_samples, = struct.unpack('I', f.read(4))
        buf = f.read(n_samples * sample_format.size)
        stride = (1 + 1 + 6) * 4
        targets = np.ndarray(n_samples, 'f4', buf, offset=0, strides=stride)
        modes_flags = np.ndarray(n_samples, 'u4', buf, offset=1*4, strides=stride)
        verts = np.ndarray((n_samples, 6), 'f4', buf, offset=(1+1)*4, strides=(stride,4))

        keep = inside_aabb_vec(cell_aabb, verts[:,v2off:v2off+3])
        configs = np.unique(modes_flags, axis=0)
        for config in configs:
            this_config = (modes_flags == config)
            this_config_keep = np.logical_and(keep, this_config)
            targets_config = targets[this_config_keep]
            verts_config = verts[this_config_keep]
            grouped_samples.append(Samples(targets_config, verts_config))
    return grouped_samples, configs

class Samples:
    def __init__(self, targets, verts):
        self.targets = targets
        self.verts = verts

class MixtureComponent:
    def is_guide_pairs(self):
        return len(self.mean) == 6

    def logpdf(self, x):
        # without weight, no restricted support
        logdet = 2. * np.sum(np.log(self.w))
        norm = -0.5 * (len(self.w) * np.log(2. * np.pi) + logdet)
        d = np.diag(1./self.w) @ self.E.T @ (x - self.mean)
        exponent = -0.5 * np.dot(d, d)
        return norm + exponent

    def logpdf_vec(self, x):
        # without weight, no restricted support
        logdet = 2. * np.sum(np.log(self.w))
        norm = -0.5 * (len(self.w) * np.log(2. * np.pi) + logdet)
        d = np.diag(1./self.w) @ self.E.T @ (x - self.mean).T
        exponent = -0.5 * np.sum(d * d, axis=0)
        return norm + exponent

    def update(self, weight, mean, cov, ess, prior_radius, restrict=True):
        self.weight = weight
        self.mean = mean
        if restrict:
            cov = restrict_cov(cov, prior_radius, ess)
        self.w, self.E = np.linalg.eigh(cov)
        self.w = np.maximum(1e-3, np.sqrt(np.maximum(0., self.w)))

        S = self.E @ np.diag(self.w**2) @ self.E.T
        if self.is_guide_pairs():
            self.w_v2, self.E_v2 = np.linalg.eigh(S[0:3, 0:3])
            self.w_v2 = np.sqrt(self.w_v2)
        else:
            self.w_v2, self.E_v2 = np.linalg.eigh(S[3:6, 3:6])
            self.w_v2 = np.sqrt(self.w_v2)

    """
    def merge(a, b):
        new_lobe = Lobe(0)
        new_lobe.octree_pos = a.octree_pos
        new_lobe.split_cost = 0
        weight = a.weight + b.weight
        delta = b.mean - a.mean
        mean = a.mean + b.weight / weight * delta
        delta_prime = b.mean - mean
        a_cov = a.E @ np.diag(a.w**2) @ a.E.T
        b_cov = b.E @ np.diag(b.w**2) @ b.E.T
        cov = a_cov * a.weight + b_cov * b.weight + b.weight * np.outer(delta, delta_prime)
        cov /= weight
        new_lobe.prior_mean = mean
        new_lobe.prior_cov = cov
        new_lobe.update(weight, mean, cov, 10000)
        return new_lobe
    """

    def to_lobe(self):
        lobe = Lobe(0)
        lobe.weight = self.weight
        lobe.modes_flags = None # TODO
        S = self.E @ np.diag(self.w**2) @ self.E.T
        lobe.mean = self.mean
        if self.is_guide_pairs():
            S_v1v2 = S[:3,:3]
            lobe.cond_mean = S[3:,:3] @ np.linalg.inv(S_v1v2)
            S_cond = S[3:,3:] - lobe.cond_mean @ S[:3,3:]
        else:
            S_v1v2 = S[:6,:6]
            lobe.cond_mean = S[6:,:6] @ np.linalg.inv(S_v1v2)
            S_cond = S[6:,6:] - lobe.cond_mean @ S[:6,6:]
        w_cond, E_cond = np.linalg.eigh(S_cond)
        lobe.E_cond = E_cond
        lobe.w_cond = np.maximum(1e-3, np.sqrt(np.maximum(0., w_cond)))
        w_v1v2, E_v1v2 = np.linalg.eigh(S_v1v2)
        lobe.E_v1v2_T = E_v1v2.T
        lobe.w_v1v2_inv = 1./np.maximum(1e-3, np.sqrt(np.maximum(0., w_v1v2)))
        return lobe

def em_fit(samples, prior_radius):
    n_components = min(max(1, len(samples.targets)), 8)
    n_components = 32
    ess = np.sum(samples.targets)**2/np.sum(samples.targets**2)
    print(n_components, len(samples.targets), ess)

    init, all_init = pca_part(samples, n_components, prior_radius)
    all_lobes = [clusters_to_comp(clusters) for clusters in all_init]

    current_mixture = clusters_to_comp(init)
    all_lobes.append(current_mixture)
    for em_iter in range(em_iterations):
        weights = eval_mixture_resp_vec(current_mixture, samples.verts)
        next_mixture = []
        for i, old_lobe in enumerate(current_mixture):
            lobe = copy.deepcopy(old_lobe)
            w = weights[:,i] * samples.targets
            if np.isclose(0, np.sum(w)):
                # dont let die completely?
                lobe.update(0, np.zeros(guide_dim), np.zeros((guide_dim,guide_dim)), 0, prior_radius, restrict=False)
                continue
            new_mean = np.average(samples.verts, axis=0, weights=w)
            new_cov = np.cov(samples.verts.T, ddof=0, aweights=w)
            new_weight = np.mean(w)
            ess = np.sum(w)**2 / np.sum(w**2)
            lobe.update(new_weight, new_mean, new_cov, ess, prior_radius)
            next_mixture.append(lobe)
        current_mixture = next_mixture
        all_lobes.append(current_mixture)
    return all_lobes

def clusters_to_comp(clusters):
    mixture = []
    for cluster in clusters:
        comp = MixtureComponent()
        comp.update(cluster.weight, cluster.mean, cluster.cov, 0, 0, restrict=False)
        mixture.append(comp)
    return mixture

def eval_mixture_resp(mixture, x):
    # https://gregorygundersen.com/blog/2020/02/09/log-sum-exp/
    logpdf = np.array([np.log(lobe.weight) + lobe.logpdf(x.verts) for lobe in mixture])
    weights = np.exp(logpdf - logsumexp(logpdf))
    return weights

def eval_mixture_resp_vec(mixture, samples):
    # https://gregorygundersen.com/blog/2020/02/09/log-sum-exp/
    logpdf = np.column_stack([np.log(lobe.weight) + lobe.logpdf_vec(samples) for lobe in mixture])
    weights = np.exp(logpdf - logsumexp_vec(logpdf)[:,np.newaxis])
    return weights

def eval_mixture_pdf_vec(mixture, samples):
    total_weight = sum([lobe.weight for lobe in mixture])
    logpdf = np.column_stack([np.log(lobe.weight/total_weight) + lobe.logpdf_vec(samples) for lobe in mixture])
    return logsumexp_vec(logpdf)

def logsumexp_vec(x):
    c = x.max(axis=1)
    return c + np.log(np.sum(np.exp(x - c[:,np.newaxis]), axis=1))

def logsumexp(x):
    c = x.max()
    return c + np.log(np.sum(np.exp(x - c)))

def pca_part(samples, k, prior_radius):
    clusters = [Cluster(np.arange(len(samples.targets)), samples, prior_radius)]
    all_clusters = [copy.deepcopy(clusters)]
    queue = [0]
    while len(clusters) < k and queue:
        split_queue_index = np.argmin([clusters[ci].log_l for ci in queue])
        split_index = queue.pop(split_queue_index)
        split_result = clusters[split_index].split(samples)
        if split_result is None:
            continue
        new_cluster_1, new_cluster_2 = split_result
        """
        bic_parent, bic_children = eval_bic(clusters[split_index], new_cluster_1, new_cluster_2, samples)
        if bic_children < bic_parent:
            continue
        print('parent: ', bic_parent, 'children: ', bic_children)
        """
        clusters[split_index] = new_cluster_1
        clusters.append(new_cluster_2)
        queue.append(split_index)
        queue.append(len(clusters) - 1)
        all_clusters.append(copy.deepcopy(clusters))
    print('total clusters: ', len(clusters))
    return clusters, all_clusters

def eval_bic(parent, child1, child2, samples):
    comps = clusters_to_comp([parent, child1, child2])
    log_l_parent = np.sum(comps[0].logpdf_vec(samples.verts[parent.sample_idx]))
    log_l_child1 = np.sum(comps[1].logpdf_vec(samples.verts[child1.sample_idx]))
    log_l_child2 = np.sum(comps[2].logpdf_vec(samples.verts[child2.sample_idx]))
    k_parent = (guide_dim * (guide_dim+1)) / 2 + guide_dim
    k_child = k_parent
    ess_parent = np.sum(samples.targets[parent.sample_idx])**2 / np.sum(samples.targets[parent.sample_idx]**2)
    ess_child1 = np.sum(samples.targets[child1.sample_idx])**2 / np.sum(samples.targets[child1.sample_idx]**2)
    ess_child2 = np.sum(samples.targets[child2.sample_idx])**2 / np.sum(samples.targets[child2.sample_idx]**2)
    bic_parent = log_l_parent - 0.5 * k_parent * np.log(ess_parent)
    bic_children = log_l_child1 + log_l_child2 - 0.5 * k_child * (np.log(ess_child1)) - 0.5 * k_child * (np.log(ess_child2)) - 0.5 * 1 * np.log(ess_parent)
    return bic_parent, bic_children
    # log likelihood = \sum log(pdf(x_i))
    # bic = log likelihood - 0.5 * k * log(N)
    # assume hard clustering here


class Cluster:
    def __init__(self, sample_idx, samples, prior_radius):
        self.sample_idx = sample_idx
        w = samples.targets[sample_idx]
        weight = np.sum(w)
        self.weight = weight
        ess = np.sum(w)**2 / np.sum(w**2)
        self.mean = np.average(samples.verts[sample_idx], axis=0, weights=w)
        cov = np.cov(samples.verts[sample_idx].T, ddof=0, aweights=w)
        self.cov = restrict_cov(cov, prior_radius, ess)
        self.prior_radius = prior_radius

        total_weight = np.sum(samples.targets)
        dim = 6
        sgn, logdet = np.linalg.slogdet(self.cov)
        self.log_l = (weight * np.log(weight / total_weight)
            - 0.5 * dim * weight * np.log(2. * np.pi)
            - 0.5 * weight * logdet
            - 0.5 * dim * weight)

    def split(self, samples):
        if(len(samples.targets) < 2):
            return None
        w, E = np.linalg.eigh(self.cov)
        max_ll = -np.inf
        k_max = -1
        for k in range(6):
            v = E[:, k]
            split_result = self.split_along(samples, v)
            if split_result is None:
                continue
            else:
                cluster1, cluster2 = split_result
            log_l = cluster1.log_l + cluster2.log_l
            if log_l > max_ll:
                max_ll = log_l
                k_max = k
        if k_max == -1:
            return None
        else:
            clusters = self.split_along(samples, E[:, k_max])
            #return clusters
            res = self.split_em(samples, clusters)
            if res is None:
                return None
            cluster1, cluster2, bic_split = res
            cluster_weights = samples.targets[self.sample_idx]
            cluster_verts = samples.verts[self.sample_idx]
            ess_total = np.sum(cluster_weights)**2 / np.sum(cluster_weights**2)
            log_l_parent = np.average(eval_mixture_pdf_vec(clusters_to_comp([self]), cluster_verts), weights=cluster_weights)
            K = (guide_dim * (guide_dim+1))/2 + guide_dim
            bic_parent = log_l_parent * ess_total - 0.5 * K * 2#np.log(ess_total)
            print("parent: ", bic_parent, "split: ", bic_split)
            if bic_split < bic_parent:
                return None
            return cluster1, cluster2

    def split_along(self, samples, v):
        alpha = np.dot(v, self.mean)
        proj = samples.verts[self.sample_idx] @ v
        idx1 = self.sample_idx[proj >= alpha]
        idx2 = self.sample_idx[proj < alpha]
        if len(idx1) < 1 or len(idx2) < 1:
            return None
        cluster1 = Cluster(idx1, samples, self.prior_radius)
        cluster2 = Cluster(idx2, samples, self.prior_radius)
        return cluster1, cluster2

    def split_em(self, samples, clusters):
        if clusters is None:
            return None
        prior_radius = self.prior_radius
        cluster_verts = samples.verts[self.sample_idx]
        cluster_weights = samples.targets[self.sample_idx]
        mixture = clusters_to_comp(clusters)
        for em_iter in range(250):
            resp = eval_mixture_resp_vec(mixture, cluster_verts)
            next_mixture = []
            for i, old_lobe in enumerate(mixture):
                lobe = copy.deepcopy(old_lobe)
                w = resp[:,i] * cluster_weights
                if np.isclose(0, np.sum(w)):
                    return None
                new_mean = np.average(cluster_verts, axis=0, weights=w)
                new_cov = np.cov(cluster_verts.T, ddof=0, aweights=w)
                new_weight = np.mean(w)
                ess = np.sum(w)**2 / np.sum(w**2)
                lobe.update(new_weight, new_mean, new_cov, ess, prior_radius)
                next_mixture.append(lobe)
            mixture = next_mixture
        weights = eval_mixture_resp_vec(mixture, cluster_verts)
        ess_total = np.sum(cluster_weights)**2 / np.sum(cluster_weights**2)
        w1 = np.sum(weights[:,0] * cluster_weights)
        w2 = np.sum(weights[:,1] * cluster_weights)
        ess1 = np.sum(w1)**2 / np.sum(w1**2)
        ess2 = np.sum(w2)**2 / np.sum(w2**2)
        N1 = mixture[0].weight
        N2 = mixture[1].weight
        wsum = N1 + N2
        N1 /= wsum
        N2 /= wsum
        N1 *= ess_total
        N2 *= ess_total
        log_l = np.average(eval_mixture_pdf_vec(mixture, cluster_verts), weights=cluster_weights)
        K = (guide_dim * (guide_dim + 1)) / 2 + guide_dim
        hbic = log_l * ess_total - 0.5 * K * (np.log(N1) + np.log(N2)) - 0.5 * 1 * np.log(ess_total)
        #N = len(cluster_weights)
        N = ess_total
        bic = log_l * N - 0.5 * (K * 2 + 1) * 2.#np.log(N)
        print("hbic:", hbic, "bic: ", bic)
        idx1 = self.sample_idx[weights[:,0] > weights[:,1]]
        idx2 = self.sample_idx[weights[:,0] <= weights[:,1]]
        if len(idx1) < 1 or len(idx2) < 1:
            return None
        cluster1 = Cluster(idx1, samples, prior_radius)
        cluster2 = Cluster(idx2, samples, prior_radius)
        return cluster1, cluster2, bic


def resample_varopt(samples, ratio=0.5):
    n = len(samples.targets)
    k_min = min(10, n)
    k = max(int(ratio * n), k_min)
    assert(np.all(samples.targets >= 1e-12))
    assert(np.all(samples.targets <= 1e9))
    L, T, tau = varopt(samples.targets, k)
    indices = [l[1] for l in L] + T
    weights = np.array([l[0] for l in L] + [tau for _ in T])
    verts = samples.verts[indices]
    return Samples(weights, verts)

def varopt(weights, k):
    n = len(weights)
    assert(n >= k)
    L = []
    T = []
    tau = 0
    for i in range(k):
        heapq.heappush(L, (weights[i], i))
    for i in range(k+1, n):
        X = []
        W = tau * len(T)
        wi = weights[i]
        if wi <= tau:
            X.append((wi, i))
            W += wi
        else:
            heapq.heappush(L, (wi, i))
        while L and W >= L[0][0] * (len(X) + len(T) - 1):
            (wj, j) = heapq.heappop(L)
            X.append((wj, j))
            W += wj
        tau = W / (len(X) + len(T) - 1)
        r = np.random.rand()
        Wx = 0
        deleted = False
        for d in range(len(X)):
            Wx += X[d][0]
            if (d + 1) - Wx / tau >= r:
                X.pop(d)
                deleted = True
                break
        if not deleted and len(T) == 0:
            print(len(X) - Wx / tau, r)
            tau = W / len(T)
        if not deleted and len(T) > 0:
            T.pop(np.random.randint(len(T)))
        for (_, idx) in X:
            T.append(idx)
    return L, T, tau

def mixture_reduce_step(mixture):
    mixture = copy.deepcopy(mixture)
    weightsum = 0
    for lobe in mixture:
        weightsum += lobe.weight
    for lobe in mixture:
        lobe.weight /= weightsum
    n = len(mixture)
    if n <= 1:
        return mixture, 0
    smallest_cost = 1e9
    merge_pair = None
    for i in range(n):
        for j in range(i+1, n):
            new_comp = Lobe.merge(mixture[i], mixture[j])
            cost = new_comp.weight * np.sum(np.log(new_comp.w)) - mixture[i].weight * np.sum(np.log(mixture[i].w)) - mixture[j].weight * np.sum(np.log(mixture[j].w))
            if cost < smallest_cost:
                smallest_cost = cost
                merge_pair = (i, j)
    if merge_pair is not None:
        i, j = merge_pair
        new_comp = Lobe.merge(mixture[i], mixture[j])
        mixture[i] = new_comp
        mixture.pop(j)
    return mixture, smallest_cost

def synthetic_samples():
    N = 7
    targets = np.ones(N)
    verts = np.zeros((N, 6))
    verts += np.random.randn(*verts.shape) * 0.5
    verts[:,0] += 10.
    verts[:,0] *= 4.
    verts[:,1] *= 1.
    verts[:,2] *= .25
    verts[:,3:6] = [0.2, .2, .02] * np.random.randn(*verts[:,3:6].shape)
    verts[:,3:6] += verts[:,0:3]
    verts[:,3:6] += 3.
    #verts[:,0:2] *= 3.
    samples = Samples(targets, verts)
    return samples

def restrict_cov(cov, prior_radius, ess):
    if ess > 1.:
        cov *= (ess/(ess-1.))
        cov = np.nan_to_num(cov)
    else:
        cov = np.zeros(cov.shape)
    cov_prior = np.eye(guide_dim) * prior_radius**2
    cov_reg = np.zeros((guide_dim, guide_dim))
    assert(guide_dim == 6)
    cov_reg[:3,:3] = np.eye(3) * np.mean(np.diag(cov[:3,:3]))
    cov_reg[3:,3:] = np.eye(3) * np.mean(np.diag(cov[3:,3:]))
    cov_reg[3:,:3] = np.eye(3) * np.mean(np.diag(cov[3:,:3]))
    for i in range(guide_dim // 3):
        for j in range(guide_dim // 3):
            c = np.mean(np.diag(cov[i*3:(i+1)*3,j*3:(j+1)*3]))
            cov_reg[i*3:(i+1)*3,j*3:(j+1)*3] = np.eye(3) * c
    t_prior = (ess - 1.) / 3.
    t_prior = np.clip(t_prior, 0., 1.)
    cov_iso = (1. - t_prior) * cov_prior + t_prior * cov_reg

    REG_N0 = guide_dim
    t_reg = REG_N0 / ((ess-1.) + REG_N0)
    t_reg = np.clip(t_reg, 0., 1.)
    t_cov = 1. - t_reg
    t_cov = np.clip(t_cov, 0., 1.)
    return cov * t_cov + cov_iso * t_reg
