import bpy
from bpy import context as C
from bpy import data as D
from mathutils import *
import struct
import numpy as np
import bmesh
import os
import importlib

import lobe
importlib.reload(lobe)
from lobe import Lobe

import lobevis
importlib.reload(lobevis)
from lobevis import LobeVis

import octree
importlib.reload(octree)
from octree import Octree

import octreevis
importlib.reload(octreevis)
from octreevis import OctreeVis

import mixture
importlib.reload(mixture)
from mixture import Mixture

#scene_dir = '/home/vincent/scenes/guiding3v/'
scene_dir = '/home/vincent/scenes/spheresbox/'
scene = 'smalllight-dielbox'
postfix = 'render'
cache_iter = 16
guide_pairs = True
show_camera = True

guide_vis = None

def main():
    global guide_vis
    assert(guide_vis is None)
    guide_vis = GuideVis()
    bpy.app.timers.register(timer_callback)

def timer_callback():
    global guide_vis
    time_to_next_call = 0.1
    if guide_vis.update():
        return time_to_next_call
    else:
        return None
    return None

query_name = 'query'

class GuideVis():
    def __init__(self):
        all_lobes, all_octrees, all_mixtures, all_cam_lobes = self.load_data()
        if show_camera:
            self.lobe_vis = LobeVis(all_cam_lobes, all_octrees, all_mixtures=None)
        else:
            self.lobe_vis = LobeVis(all_lobes, all_octrees, all_mixtures, show_all=True)
        self.octree_vis = OctreeVis(all_octrees)
        self.make_query_object()
        self.frame = None

    def load_data(self):
        all_lobes = []
        all_octrees = []
        all_mixtures = []
        all_cam_lobes = []
        for i in range(1, cache_iter+1):
            cache_path = (scene_dir
                + scene
                #+'/corona/'
                #+ scene
                + postfix
                + '_cache{:05d}_vis'.format(i))
            with open(cache_path, 'rb') as f:
                lobes = Lobe.read_from_file(f)
                octree = Octree.read_from_file(f)
                mixtures = Mixture.read_from_file(f)
                cam_lobes = Lobe.read_from_file_camera(f)
            all_lobes.append(lobes)
            all_octrees.append(octree)
            all_mixtures.append(mixtures)
            all_cam_lobes.append(cam_lobes)
        return all_lobes, all_octrees, all_mixtures, all_cam_lobes

    def make_query_object(self):
        if query_name in D.objects:
            D.objects.remove(D.objects[query_name])
        bpy.ops.object.empty_add(type='PLAIN_AXES')
        query_obj = C.active_object
        query_obj.name = query_name
        query_obj.empty_display_size = 10
        self.query_pos = None

    def update(self):
        if query_name not in D.objects:
            print("stopping")
            return False
        query_obj = D.objects[query_name]
        new_frame = C.scene.frame_current
        frame_did_change = new_frame != self.frame
        pos_did_change = query_obj.location != self.query_pos
        self.query_pos = Vector(query_obj.location)
        self.frame = new_frame

        if frame_did_change:
            self.octree_vis.update(self.frame)

        if frame_did_change or pos_did_change:
            self.lobe_vis.update(self.frame, np.array(self.query_pos))

        return True

if __name__ == "__main__":
    main()
