import struct
import numpy as np

class Octree:
    def __init__(self, aabb, nodes):
        self.aabb = aabb
        self.nodes = nodes

    def read_from_file(f):
        node_format = struct.Struct('=I')
        n_nodes, = struct.unpack('=I', f.read(4))
        aabb = np.array(struct.unpack('=6f', f.read(4 * 6)))
        buf = f.read(n_nodes * node_format.size)
        nodes_raw = node_format.iter_unpack(buf)
        nodes = [x[0] for x in nodes_raw]
        return Octree(aabb, nodes)

    def is_leaf(node):
        return bool(node & (1<<0))

    def node_index(node):
        return node >> 1

    def get_child_offset(aabb, pos):
        offset = 0
        for k in range(3):
            split = aabb[k] + 0.5 * (aabb[3+k] - aabb[k])
            offset |= (1 if split <= pos[k] else 0) << k
        return offset

    def compute_child_aabb(aabb, offset):
        for k in range(3):
            split = aabb[k] + 0.5 * (aabb[3+k] - aabb[k])
            if offset & (1<<k):
                aabb[k] = split
            else:
                aabb[3+k] = split

    def lookup_index(self, pos):
        cell_aabb = self.aabb.copy()
        index = 0
        while True:
            child_offset = Octree.get_child_offset(cell_aabb, pos)
            Octree.compute_child_aabb(cell_aabb, child_offset)
            child_index = Octree.node_index(index) + child_offset
            index = self.nodes[child_index]
            if Octree.is_leaf(index):
                return Octree.node_index(index)

    def lookup_aabb(self, pos):
        cell_aabb = self.aabb.copy()
        index = 0
        while True:
            child_offset = Octree.get_child_offset(cell_aabb, pos)
            Octree.compute_child_aabb(cell_aabb, child_offset)
            child_index = Octree.node_index(index) + child_offset
            index = self.nodes[child_index]
            if Octree.is_leaf(index):
                return cell_aabb
