import bpy
from bpy import data as D
import bmesh

import numpy as np

import importlib
import octree
importlib.reload(octree)
from octree import Octree

class OctreeVis:
    def __init__(self, all_octrees):
        octree_name = 'octree'
        if octree_name in D.meshes:
            D.meshes.remove(D.meshes[octree_name])
        mesh = bpy.data.meshes.new(octree_name)
        if octree_name in D.objects:
            D.objects.remove(D.objects[octree_name])
        self.octree_obj = bpy.data.objects.new(octree_name, mesh)
        bpy.context.scene.collection.objects.link(self.octree_obj)
        self.all_octrees = all_octrees

    def update(self, frame):
        frame = np.clip(frame, 1, len(self.all_octrees)) - 1
        octree = self.all_octrees[frame]
        aabbs = [octree.aabb]
        stack = [(0, octree.aabb, 0)]
        while stack:
            node, aabb, depth = stack.pop()
            for k in range(8):
                child_aabb = aabb.copy()
                Octree.compute_child_aabb(child_aabb, k)
                aabbs.append(child_aabb)
                child_node = octree.nodes[node + k]
                if not Octree.is_leaf(child_node):
                    stack.append((Octree.node_index(child_node), child_aabb, depth + 1))

        bm = bmesh.new()
        bm.from_mesh(self.octree_obj.data)
        bm.clear()
        for aabb in aabbs:
            OctreeVis.add_aabb_vertices(bm, aabb)
        bm.verts.ensure_lookup_table()
        for i, aabb in enumerate(aabbs):
            OctreeVis.add_aabb_edges(bm, aabb, i * 8)
        bm.edges.ensure_lookup_table()
        bm.to_mesh(self.octree_obj.data)
        bm.free()

    def add_aabb_vertices(bm, aabb):
        for k in range(8):
            bm.verts.new((
                aabb[0 + (3 if k & (1<<0) else 0)],
                aabb[1 + (3 if k & (1<<1) else 0)],
                aabb[2 + (3 if k & (1<<2) else 0)]
            ))

    def add_aabb_edges(bm, aabb, offset):
        for a in range(8):
            for i in range(3):
                b = a ^ (1<<i)
                if a < b:
                    bm.edges.new((bm.verts[offset + a], bm.verts[offset + b]))
